#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: helpers.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Import all parts from helpers here.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
"""
import logging
import threading
from time import sleep
from powermolelib import Configuration
from powermolelib.powermolelibexceptions import InvalidConfigurationFile
from powermolegui.lib.logging import LoggerMixin, LOGGER_BASENAME as ROOT_LOGGER_BASENAME
from powermolegui.lib.items import ClientCanvasItem, HostCanvasItem, ConnectionCanvasItem, AgentCanvasItem, \
    PacketCanvasItem, StatusBannerCanvasItem
from powermolegui.powermoleguiexceptions import SetupFailed

__author__ = '''Vincent Schouten <powermole@protonmail.com>'''
__docformat__ = '''google'''
__date__ = '''08-10-2020'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<powermole@protonmail.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging.
logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')
LOGGER = logging.getLogger(f'{ROOT_LOGGER_BASENAME}.helpers')  # non-class objects like fn will consult this object


def parse_configuration_file(config_file_path):
    """Parses the configuration file to a (dictionary) object."""
    try:
        configuration = Configuration(config_file_path)
    except InvalidConfigurationFile:
        return None
    if configuration.mode == 'FOR':
        LOGGER.info('mode FOR enabled')
    elif configuration.mode == 'TOR':
        LOGGER.info('mode TOR enabled')
    elif configuration.mode == 'PLAIN':
        LOGGER.info('mode PLAIN enabled')
    return configuration


class ItemsGenerator(LoggerMixin):
    """Creates items for the components Client, Connection(s), Host(s) and Agent."""

    def __init__(self, main_window, configuration):
        """Instantiates the ShapeGenerator object."""
        super().__init__()
        # self.scale = main_window.scale  # IN DEVELOPMENT
        self._main_window = main_window
        self._config = configuration
        self._status = main_window.main_frame.canvas_frame.canvas_status
        self._canvas = main_window.main_frame.canvas_frame.canvas_landscape
        self._iteration = 0
        self._quit = False
        self._wipe_canvas()

    def _wipe_canvas(self):
        self._canvas.delete("all")

    def create_canvas_items(self):
        """Create all items and hide.

        The number of host items are derived on the total amount of hosts.
        The ConnectionCanvasItem needs not-hidden items to create a connection_canvas_item as it uses bbox.
        Bbox cannot work with hidden items.
        Once all canvas items are created, they are hidden.

        Returns:
            A list containing items. Each type of item has its own position:
            • element 0: client - the Client is given an unique position on the X-axis.
            • element 1: [hosts] - each Host is given an unique position on the X-axis.
            • element 2: [connection] - each Connection is given its from and to destination.
            • element 3: agent - an Agent is a 'child' of the target destination host.

        """
        start_y_pos = 60
        start_x_pos = 70
        amount_hosts = len(self._config.gateways) + 1
        client_item = ClientCanvasItem(self._main_window, start_x_pos, start_y_pos)
        client_item.create()
        host_items = []
        connection_items = []
        for i in range(amount_hosts):
            start_x_pos += 220
            host_item = HostCanvasItem(self._main_window, start_x_pos, start_y_pos, self._config.all_host_addr[i])
            host_item.create()
            host_items.append(host_item)
            if i == 0:
                connection_item = ConnectionCanvasItem(self._main_window, client_item, host_items[i])
            else:
                connection_item = ConnectionCanvasItem(self._main_window, host_items[i - 1], host_items[i])
            connection_item.create()
            connection_items.append(connection_item)
        agent_item = AgentCanvasItem(self._main_window, client_item, host_items)
        agent_item.create()
        packet_item = PacketCanvasItem(self._main_window, connection_items)
        packet_item.create()
        status_item = StatusBannerCanvasItem(self._main_window)
        status_item.create()
        self._status.itemconfig('all', state='hidden')
        self._canvas.itemconfig('all', state='hidden')
        return client_item, host_items, connection_items, agent_item, packet_item, status_item

    def show_landscape(self, canvas_items):  # pylint: disable=no-self-use
        """Shows the shapes of Client and Host(s).

        Modifying the scroll region only works after the items
        have been changed ("configured") from hidden to normal.

        """
        client_item, host_items, _, _, _, status_item = canvas_items
        sleep(0.5)  # otherwise the drawing begins earlier than the eye notices
        client_item.show()  # show client item
        for host in host_items:  # show host items
            host.show()
            sleep(0.1)  # to allow this host to finish flickering before the next host is shown
        status_item.show()


class ClientAdapter:
    """Adapts a client object to a representational state for the GUI."""

    def __init__(self, item):
        """Instantiates the ClientAdapter object."""
        self.item = item

    def __str__(self):
        return 'Client'

    def stop(self):
        """Stops."""
        self.item.dim()
        return True


class TunnelAdapter:
    """Adapts a tunnel object to a representational state for the GUI."""

    def __init__(self, object_, items):
        """Instantiates the TunnelAdapter object."""
        self.object_ = object_
        self.items = items

    def __str__(self):
        return 'Tunnel'

    def stop(self):
        """Stops."""
        for item in self.items:
            item.hide()
        return self.object_.stop()


class HostAdapter:
    """Adapts a host object to a representational state for the GUI."""

    def __init__(self, items):
        """Instantiates the HostAdapter object."""
        self.items = items

    def __str__(self):
        return 'Host'

    def stop(self):
        """Stops."""
        for host in self.items:
            host.dim()
        return True


class AgentAdapter:
    """Adapts an agent object to a representational state for the GUI."""

    def __init__(self, item):
        """Instantiates the AgentAdapter object."""
        self.item = item

    def __str__(self):
        return 'Agent'

    def stop(self):
        """Stops."""
        self.item.dim()
        return True


class SetupLink(LoggerMixin):  # pylint: disable=too-many-instance-attributes
    """Establishes a connection to target destination host via intermediaries by starting various objects.

    This function also passes the instantiated objects to the StateManager, which
    will stop the Tunnel and Instructor after a KeyboardInterrupt.
    """

    def __init__(self, state, transfer_agent, tunnel, bootstrap_agent, instructor,  # pylint: disable=too-many-arguments
                 client_item, host_items, agent_item, connection_items):
        """Instantiates the SetupLink object.

        Args:
            state (StateManager): An instantiated StateManager object.
            transfer_agent (TransferAgent): An instantiated TransferAgent object.
            tunnel (Tunnel): An instantiated Tunnel object.
            bootstrap_agent (BootstrapAgent): ...
            instructor (Instructor): ...
            client_item (ClientCanvasItem): ...
            host_items (HostCanvasItem): ...
            agent_item (AgentCanvasItem): ...
            connection_items (ConnectionCanvasItem): ...

        """
        super().__init__()
        self._is_terminate_query_scp = False
        self._is_terminate_query_ssh = False
        self._state = state
        self._transfer_agent = transfer_agent
        self._tunnel = tunnel
        self._bootstrap_agent = bootstrap_agent
        self._instructor = instructor
        self._client_item = client_item
        self._host_items = host_items
        self._agent_item = agent_item
        self._connection_items = connection_items

    def start(self):
        """Starts setting up link."""
        self.start_client()
        self.start_transfer_agent()
        self.start_tunnel()
        self.start_bootstrap_agent()
        self.start_instructor()

    def _query_transfer_agent_connection(self):
        """Moves the Agent canvas item every time the tunnel.authenticated_hosts is appended with a new host."""
        self._logger.debug('querying for authenticated hosts...')
        index = 0
        while index < len(self._transfer_agent.all_host_addr):
            # show 1st move
            if index == 0 and self._transfer_agent.all_host_addr[index] in self._transfer_agent.authenticated_hosts:
                self._agent_item.move()  # blocking
                index += 1
            # show nth move
            elif index > 0 and self._transfer_agent.all_host_addr[index] in self._transfer_agent.authenticated_hosts:
                self._agent_item.move()  # blocking
                index += 1
            if self._is_terminate_query_scp:
                break
            sleep(0.1)  # otherwise, tk might behaves erratic

    def _query_ssh_proxyjump_connection(self):
        """Shows a new Connection canvas item every time the tunnel.authenticated_hosts is appended with a new host."""
        self._logger.debug('querying for authenticated hosts...')
        index = 0
        while index < len(self._tunnel.all_host_addr):
            if index == 0 and self._tunnel.all_host_addr[index] in self._tunnel.authenticated_hosts:  # show 1st conn.
                self._connection_items[index].show()  # blocking
                index += 1
            elif index > 0 and self._tunnel.all_host_addr[index] in self._tunnel.authenticated_hosts:  # show nth conn.
                self._connection_items[index].show()  # blocking
                index += 1
            if self._is_terminate_query_ssh:
                break
            sleep(0.1)  # otherwise, tk might behaves erratic

    def start_client(self):
        """Shows the client item.

        This method also adds the instance of the client adapter to the context manager.

        """
        self._state.add_object(ClientAdapter(self._client_item))
        self._client_item.setup_ok()

    def start_transfer_agent(self):
        """Starts transferring Agent to destination host."""
        # the TransferAgent object is a disposable one-trick pony
        # no need to invoke state.add_object()
        self._agent_item.show()
        thread = threading.Thread(target=self._query_transfer_agent_connection)
        thread.start()
        if not self._transfer_agent.start():
            self._agent_item.transfer_nok()
            self._is_terminate_query_scp = True
            raise SetupFailed(self._transfer_agent)
        while thread.is_alive():
            sleep(0.5)
        self._logger.info('Agent has been transferred securely to destination host')
        self._agent_item.transfer_ok()
        sleep(0.25)  # give the user some time to follow all activity on the canvas

    def start_tunnel(self):
        """Starts setting up the Tunnel with forwarded connections used by Instructor."""
        self._state.add_object(TunnelAdapter(self._tunnel, self._connection_items))
        self._state.add_object(HostAdapter(self._host_items))
        thread = threading.Thread(target=self._query_ssh_proxyjump_connection)
        thread.start()
        if not self._tunnel.start():
            for conn, host in zip(self._connection_items, self._host_items):
                conn.setup_nok()
                host.setup_nok()
            self._is_terminate_query_ssh = True
            raise SetupFailed(self._tunnel)
        self._logger.info('Tunnel has been opened...')
        while thread.is_alive():
            sleep(0.5)
        for conn, host in zip(self._connection_items, self._host_items):
            sleep(0.1)
            conn.setup_ok()
            host.setup_ok()
        sleep(0.25)  # give the user some time to follow all activity on the canvas

    def start_bootstrap_agent(self):
        """Starts bootstrapping the agent by executing agent module on destination host."""
        # the BootstrapAgent object is a disposable one-trick pony
        if not self._bootstrap_agent.start():
            self._agent_item.setup_nok()
            raise SetupFailed(self._bootstrap_agent)
        self._logger.info('Agent has been executed')
        self._agent_item.setup_ok()

    def start_instructor(self):
        """Starts the Instructor and checks if the Agent can be reached."""
        self._state.add_object(self._instructor)
        self._state.add_object(AgentAdapter(self._agent_item))
        if not self._instructor.start():
            raise SetupFailed(self._instructor)
        self._logger.info('Instructor has been executed')


class StateVisualiser:  # pylint: disable=too-few-public-methods
    """Shows the state of the encrypted tunnel by a banner and by visualising the packet flow."""

    def __init__(self, heartbeat, animated_packet, status_item):
        """Instantiates the StateVisualiser object.

        Args:
            animated_packet: ...
            heartbeat: An instantiated Heartbeat context manager object which determines periodically the
                        state of the tunnel.
            status_item: ...

        """
        self._animated_packet = animated_packet
        self._heartbeat = heartbeat
        self._status_item = status_item

    def start(self):
        """Start continuously determining the state of the tunnel."""
        self._status_item.show('opened')
        self._animated_packet.start()
        while not self._heartbeat.terminate:
            if not self._heartbeat.is_tunnel_intact:
                self._status_item.show('broken')
                self._animated_packet.pause()
                while not self._heartbeat.is_tunnel_intact:
                    sleep(0.2)
                self._status_item.show('restored')
                self._animated_packet.resume()
            sleep(0.2)
        self._animated_packet.stop()
        self._status_item.dim()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: items.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for various tkinter canvas items.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import uuid
import re
from time import sleep
from itertools import cycle
from abc import ABC, abstractmethod

__author__ = '''Vincent Schouten <powermole@protonmail.com>'''
__docformat__ = '''google'''
__date__ = '''06-12-2019'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<powermole@protonmail.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# Constants
FONT_SIZE = 10
THICKNESS_LINE_CLIENT = 2
THICKNESS_LINE_HOST = 2
THICKNESS_LINE_AGENT = 2
THICKNESS_LINE_TUNNEL = 1
THICKNESS_LINE_PACKET = 1
THICKNESS_PACKET = 2
WIDTH_PACKET = 20
BACKGROUND_CANVAS = '#232729'
LABEL_FONT_COLOUR = 'white'
PACKET_OUTLINE_COLOUR = '#c0c0c0'
PACKET_FILL_COLOUR = '#232729'
NON_OPERATION = 'white'
OK_COLOUR = 'green'
NOK_COLOUR = 'red'


class CanvasItem(ABC):
    """Enforces methods to be implemented for the subclassed objects."""

    def __init__(self, main_window):
        """Instantiates the CanvasItem object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.

        """
        # self.scale = main_window.scale  # IN DEVELOPMENT
        self.item_name = self._determine_item_name()
        self.item_tag = f"{self.item_name}-{self._uuid}"
        self._main_window = main_window
        self._canvas_landscape = main_window.main_frame.canvas_frame.canvas_landscape
        self._canvas_status = main_window.main_frame.canvas_frame.canvas_status
        self._item_label = 0

    @property
    def _uuid(self):
        uuid_ = f'{uuid.uuid4().hex}'
        return uuid_

    def _determine_item_name(self):
        return re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', type(self).__name__)[0]  # split words at CamelCase

    def _add_label(self, item_name):
        coord_x1, coord_y1, coord_x2, _ = self._canvas_landscape.coords(self.item_tag)
        distance = coord_x2 - coord_x1
        pos_x = coord_x1 + (distance / 2)  # to determine the coordinates of the center of the shape (host/client)
        pos_y = coord_y1 - 10  # put the label a few pixels above the shape
        self._item_label = self._canvas_landscape.create_text(pos_x,
                                                              pos_y,
                                                              text=item_name,
                                                              font=('', FONT_SIZE, 'normal'),
                                                              fill=LABEL_FONT_COLOUR)

    @abstractmethod
    def create(self):
        """Creates the canvas item."""

    @abstractmethod
    def show(self):
        """Shows the canvas item."""

    @abstractmethod
    def hide(self):
        """Hides the canvas item."""

    @abstractmethod
    def setup_ok(self):
        """Colours the canvas item in accordance with an OK state."""

    @abstractmethod
    def setup_nok(self):
        """Colours the canvas item in accordance with an NOK state."""

    @abstractmethod
    def dim(self):
        """Colours the canvas item representing a non-operational state."""


class Effect:
    """Applies effects to the canvas item.

    This class provides methods to colour the canvas item in accordance with the
    operational state and provide a method to make the canvas item flicker.

    """

    def __init__(self, main_window, canvas_item, filling_type):
        """Instantiates the Effect object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.
            canvas_item (str): The ID of the canvas item on the canvas.
            filling_type (str): Either 'fill' or 'outline' depending what part of the canvas item should be coloured.

        """
        self._canvas_landscape = main_window.main_frame.canvas_frame.canvas_landscape
        self._canvas_status = main_window.main_frame.canvas_frame.canvas_status
        self._flicker_index = 0
        self._canvas_item = canvas_item
        self._filling_type = filling_type

    def flicker(self):
        """Changes the brightness of the canvas item irregularly appearing as a fluctuating light.

        The after function is threaded, so it doesn't block.

        """
        self._canvas_landscape.itemconfig(self._canvas_item, state='normal')
        colours = ['#b2b2b2', '#b2b2b2', '#7f7f7f', '#b2b2b2', 'white']
        colour_cycle = cycle(colours)
        max_elements = len(colours)

        def _flicker():
            self._flicker_index += 1
            selected_colour = next(colour_cycle)
            if self._filling_type == 'outline':
                self._canvas_landscape.itemconfig(self._canvas_item, outline=selected_colour)
                self._canvas_status.itemconfig(self._canvas_item, outline=selected_colour)
            elif self._filling_type == 'fill':
                self._canvas_landscape.itemconfig(self._canvas_item, fill=selected_colour)
                self._canvas_status.itemconfig(self._canvas_item, fill=selected_colour)
            if self._flicker_index == max_elements:
                return
            self._canvas_landscape.after(120, _flicker)
        _flicker()

    def setup_ok(self):
        """Colours the canvas item to state OK (green)."""
        arguments = {self._filling_type: OK_COLOUR}
        self._canvas_landscape.itemconfig(self._canvas_item, **arguments)
        self._canvas_status.itemconfig(self._canvas_item, **arguments)

    def setup_nok(self):
        """Colours the canvas item to state NOK (red)."""
        arguments = {self._filling_type: NOK_COLOUR}
        self._canvas_landscape.itemconfig(self._canvas_item, **arguments)
        self._canvas_status.itemconfig(self._canvas_item, **arguments)

    def dim(self):
        """Colours the canvas item to a non-operational state (white).

        To colour the outline of 'rectangles', the outline has to be configured.
        To colour the items made of 'lines', the fill has to be configured.
        """
        arguments = {self._filling_type: NON_OPERATION}
        self._canvas_landscape.itemconfig(self._canvas_item, **arguments)
        self._canvas_status.itemconfig(self._canvas_item, **arguments)


class ConnectionCalculator:
    """Calculates all properties that is needed to render a connection between two canvas items."""

    def __init__(self, main_window, canvas_item_1, canvas_item_2):
        """Instantiates the ConnectionCalculator object.

        Note: canvas.bbox doesn't return values when the item's state is 'hidden'.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.
            canvas_item_1 (CanvasItem): An instantiated CanvasItem object.
            canvas_item_2 (CanvasItem): An instantiated CanvasItem object.

        """
        self._canvas_landscape = main_window.main_frame.canvas_frame.canvas_landscape
        self._component_a = self._canvas_landscape.bbox(canvas_item_1.item_tag)
        self._component_b = self._canvas_landscape.bbox(canvas_item_2.item_tag)
        self._factor = 0.05  # why is this?

    def get_connection_length_inner(self):
        """Returns the distance between two items starting from right side first item to left side second item."""
        _, _, ax2, _ = self._component_a
        bx1, _, _, _ = self._component_b
        distance = bx1 - ax2
        return distance

    def get_connection_length_outer(self):
        """Returns the distance between two items starting from right side first item to left side second item."""
        ax1, _, _, _ = self._component_a
        _, _, bx2, _ = self._component_b
        distance = bx2 - ax1
        return distance

    def get_connection_height(self):
        """Returns the height of a connection based on the height of the item."""
        _, by1, _, by2 = self._component_b
        return (by2 - by1) * self._factor  # use bbox to return the bounding box for client & host == more sophisticated

    def get_x_pos_connection_right_side(self):
        """Returns the starting position on the X-axis of the connection item."""
        _, _, ax2, _ = self._component_a
        return ax2

    def get_x_pos_connection_left_side(self):
        """Returns the starting position on the X-axis of the connection item."""
        ax1, _, _, _ = self._component_a
        return ax1

    def get_y_pos_connection(self):
        """Returns the first position on the Y-axis of the connection item."""
        _, by1, _, by2 = self._component_b
        return by1 + ((by2 - by1) / 2)


class ClientCanvasItem(CanvasItem):
    """Creates a canvas item representing a client."""

    def __init__(self, main_window, start_pos_x, start_pos_y):
        """Instantiates the ClientCanvasItem object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.
            start_pos_x (int): The position on the X-axis of the canvas item's top left corner.
            start_pos_y (int): The position on the Y-axis of the canvas item's top left corner.

        """
        super().__init__(main_window)
        self._client_effect = Effect(main_window, self.item_tag, 'outline')
        self._start_pos_x = start_pos_x
        self._start_pos_y = start_pos_y
        self._width = 90
        self._height = self._width * 0.7
        self._components = []

    def create(self):
        self._draw_outer_screen()
        self._draw_inner_screen()
        self._draw_keyboard()
        self._draw_spacebar()

    def show(self):
        self._client_effect.flicker()
        self._add_label(self.item_name)

    def hide(self):
        pass

    def _draw_outer_screen(self):
        component = self._canvas_landscape.create_rectangle(self._start_pos_x,
                                                            self._start_pos_y,
                                                            self._start_pos_x + self._width,
                                                            self._start_pos_y + self._height,
                                                            width=THICKNESS_LINE_CLIENT,
                                                            outline=NON_OPERATION,
                                                            tags=self.item_tag
                                                            # state='hidden'
                                                            )
        self._components.append(component)

    def _draw_inner_screen(self):
        coord_x1, coord_y1, coord_x2, coord_y2 = self._canvas_landscape.coords(self._components[0])
        total_width = coord_x2 - coord_x1
        total_height = coord_y2 - coord_y1
        width = (coord_x2 - coord_x1) * 0.9
        height = (coord_y2 - coord_y1) * 0.9
        start_pos_x = coord_x1 + ((total_width - width) / 2)
        start_pos_y = coord_y1 + ((total_height - height) / 2)
        component = self._canvas_landscape.create_rectangle(start_pos_x,
                                                            start_pos_y,
                                                            start_pos_x + width,
                                                            start_pos_y + height,
                                                            width=1,
                                                            outline=NON_OPERATION,
                                                            tags=self.item_tag
                                                            # state='hidden'
                                                            )
        self._components.append(component)

    def _draw_keyboard(self):
        coord_x1, coord_y1, coord_x2, coord_y2 = self._canvas_landscape.coords(self._components[0])
        width = coord_x2 - coord_x1
        total_height = coord_y2 - coord_y1
        height = (coord_y2 - coord_y1) * 0.7
        start_pos_x = coord_x1
        start_pos_y = coord_y2 + (total_height * 0.05)  # value should not be hardcoded, but calculated
        component = self._canvas_landscape.create_rectangle(start_pos_x,
                                                            start_pos_y,
                                                            start_pos_x + width,
                                                            start_pos_y + height,
                                                            width=THICKNESS_LINE_CLIENT,
                                                            outline=NON_OPERATION,
                                                            tags=self.item_tag
                                                            # state='hidden'
                                                            )
        self._components.append(component)

    def _draw_spacebar(self):
        coord_x1, coord_y1, coord_x2, coord_y2 = self._canvas_landscape.coords(self._components[2])
        total_width = coord_x2 - coord_x1
        total_height = coord_y2 - coord_y1
        width = total_width * 0.5
        height = total_height * 0.2
        start_pos_x = coord_x1 + ((total_width - width) * 0.5)
        start_pos_y = coord_y1 + ((total_height - height) * 0.8)
        component = self._canvas_landscape.create_rectangle(start_pos_x,
                                                            start_pos_y,
                                                            start_pos_x + width,
                                                            start_pos_y + height,
                                                            fill='',
                                                            width=1,
                                                            outline=NON_OPERATION,
                                                            tags=self.item_tag
                                                            # state='hidden'
                                                            )
        self._components.append(component)

    def setup_ok(self):
        self._client_effect.setup_ok()

    def setup_nok(self):
        self._client_effect.setup_nok()

    def dim(self):
        self._client_effect.dim()


class HostCanvasItem(CanvasItem):
    """Creates a canvas item representing a host."""

    def __init__(self, main_window, start_pos_x, start_pos_y, host_ip):
        """Instantiates the HostCanvasItem object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.
            start_pos_x (int): The position on the X-axis of the canvas item's top left corner.
            start_pos_y (int): The position on the Y-axis of the canvas item's top left corner.
            host_ip (str): The IP address of the Host.

        """
        super().__init__(main_window)
        self._host_effect = Effect(main_window, self.item_tag, 'outline')
        self._start_pos_x = start_pos_x
        self._start_pos_y = start_pos_y
        self._width = 90
        self._height = self._width * 1.4
        self._host_ip = host_ip

    def create(self):
        host_x1 = self._start_pos_x
        host_y1 = self._start_pos_y
        host_x2 = self._start_pos_x + self._width
        host_y2 = self._start_pos_y + self._height
        self._canvas_landscape.create_rectangle(host_x1,
                                                host_y1,
                                                host_x2,
                                                host_y2,
                                                width=THICKNESS_LINE_HOST,
                                                outline=NON_OPERATION,
                                                tags=self.item_tag,
                                                # state='hidden'
                                                )

    def show(self):
        self._host_effect.flicker()
        self._add_label(self._host_ip)

    def hide(self):
        pass

    def setup_ok(self):
        self._host_effect.setup_ok()

    def setup_nok(self):
        self._host_effect.setup_nok()

    def dim(self):
        self._host_effect.dim()


class AgentCanvasItem(CanvasItem):  # pylint: disable=too-many-instance-attributes
    """Creates a canvas item representing an Agent."""

    def __init__(self, main_window, client_canvas_item, host_canvas_items):
        """Instantiates the AgentCanvasItem object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.
            client_canvas_item (CanvasItem): An instantiated CanvasItem that represents a Client.
            host_canvas_items (list): All instantiated CanvasItems that represents a Host.

        """
        super().__init__(main_window)
        self._agent_effect = Effect(main_window, self.item_tag, 'outline')
        self._coords_client = self._canvas_landscape.coords(client_canvas_item.item_tag)
        self._coords_destination_host = self._canvas_landscape.coords(host_canvas_items[-1].item_tag)
        self._coords_agent_start_pos = None  # position of Agent next to Client
        self._coords_agent_end_pos = None  # position of Agent inside Host
        self._number_of_hosts = len(host_canvas_items)
        self._width = 0
        self._distance_index = 0
        self._distance_between_hosts = 0

    def create(self):
        self._coords_agent_end_pos = self._create_agent_derived_from_host()
        self._width = self._determine_width()
        self._coords_agent_start_pos = self._place_agent_in_client()
        self._distance_between_hosts = self._calculate_distances()

    def show(self):
        self._canvas_landscape.itemconfig(self.item_tag, dash=(9, 9), width=1, state='normal')

    def hide(self):
        pass

    def _create_agent_derived_from_host(self):
        hx1, hy1, hx2, hy2 = self._coords_destination_host
        nx1 = hx1 + 25
        ny1 = hy1 + 45
        nx2 = hx2 - 25
        ny2 = hy2 - 25
        self._canvas_landscape.create_rectangle(nx1,
                                                ny1,
                                                nx2,
                                                ny2,
                                                fill=BACKGROUND_CANVAS,
                                                width=THICKNESS_LINE_AGENT,
                                                outline=NON_OPERATION,
                                                tags=self.item_tag,
                                                # state='hidden'
                                                )
        return nx1, ny1, nx2, ny2

    def _determine_width(self):
        ax1, _, ax2, _ = self._canvas_landscape.coords(self.item_tag)
        return ax2 - ax1

    def _place_agent_in_client(self):
        _, ay1, _, ay2 = self._coords_agent_end_pos
        cx1, _, cx2, _ = self._coords_client
        center = cx1 + ((cx2 - cx1) / 2)
        nx1 = center - (self._width / 2)
        nx2 = center + (self._width / 2)
        self._canvas_landscape.coords(self.item_tag,
                                      nx1,
                                      ay1,
                                      nx2,
                                      ay2)
        return nx1, ay1, nx2, ay2

    def _calculate_distances(self):
        start_x2 = self._coords_agent_start_pos[0]  # retrieve top left position on x-axis
        end_x1 = self._coords_agent_end_pos[2]  # retrieve right bottom position on x-axis
        return (end_x1 - start_x2 - self._width) / self._number_of_hosts

    def move(self):
        """Moves the Agent item from Client item to destination Host item."""
        def _animate():
            stepper = self._distance_index / 10
            while self._distance_index > 0:
                self._canvas_landscape.move(self.item_tag, stepper, 0)
                self._distance_index -= stepper
                sleep(.02)

        self._distance_index = self._distance_between_hosts
        _animate()

    def transfer_ok(self):
        """Colours the outline white and changes the outline into a solid pattern."""
        self._canvas_landscape.itemconfig(self.item_tag,
                                          dash=(1, 1),  # default to normal, it was dash=(5, 5)
                                          outline=NON_OPERATION,
                                          width=THICKNESS_LINE_AGENT)
        self._add_label(self.item_name)

    def transfer_nok(self):
        """Colours the outline red and changes the outline to have a dashed pattern."""
        self._canvas_landscape.itemconfig(self.item_tag, dash=(5, 5), outline=NOK_COLOUR, width=THICKNESS_LINE_AGENT)

    def setup_ok(self):
        self._agent_effect.setup_ok()

    def setup_nok(self):
        self._agent_effect.setup_nok()

    def dim(self):
        self._agent_effect.dim()


class ConnectionCanvasItem(CanvasItem):  # pylint: disable=too-many-instance-attributes
    """Creates a canvas item representing a connection."""

    def __init__(self, main_window, canvas_item_1, canvas_item_2):
        """Instantiates the ConnectionCanvasItem object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.
            canvas_item_1 (CanvasItem): An instantiated CanvasItem object, representing either an Agent or a Host.
            canvas_item_2 (CanvasItem): An instantiated CanvasItem object, representing a Host.

        """
        super().__init__(main_window)
        self._connection_calculator = ConnectionCalculator(main_window, canvas_item_1, canvas_item_2)  # composition
        self._connection_effect = Effect(main_window, self.item_tag, 'fill')
        self._start_pos_x = self._connection_calculator.get_x_pos_connection_right_side()
        self._pos_y_1 = self._connection_calculator.get_y_pos_connection() - 8
        self._pos_y_2 = self._connection_calculator.get_y_pos_connection() + 8
        self._distance = self._connection_calculator.get_connection_length_inner()
        self._terminate = False
        self._distance_index = 0
        self._components = []

    def create(self):
        top_line = self._canvas_landscape.create_line(self._start_pos_x,
                                                      self._pos_y_1,
                                                      self._start_pos_x + self._distance,
                                                      self._pos_y_1,
                                                      fill=NON_OPERATION,
                                                      width=THICKNESS_LINE_TUNNEL,
                                                      tags=self.item_tag,
                                                      # state='hidden'
                                                      )
        self._components.append(top_line)
        bottom_line = self._canvas_landscape.create_line(self._start_pos_x,
                                                         self._pos_y_2,
                                                         self._start_pos_x + self._distance,
                                                         self._pos_y_2,
                                                         fill=NON_OPERATION,
                                                         width=THICKNESS_LINE_TUNNEL,
                                                         tags=self.item_tag,
                                                         # state='hidden'
                                                         )
        self._components.append(bottom_line)

    def show(self):
        self._canvas_landscape.itemconfig(self.item_tag, state='normal')

        def animate():  # blocking
            self._distance_index = self._distance
            pos_x_stepper = self._start_pos_x
            stepper = self._distance / 8
            while self._distance_index >= 0:
                self._canvas_landscape.coords(self._components[0],
                                              self._start_pos_x,
                                              self._pos_y_1,
                                              pos_x_stepper,
                                              self._pos_y_1)
                self._canvas_landscape.coords(self._components[1],
                                              self._start_pos_x,
                                              self._pos_y_2,
                                              pos_x_stepper,
                                              self._pos_y_2)
                self._distance_index -= stepper
                pos_x_stepper += stepper
                sleep(.01)
            self._add_label(self.item_name)

        animate()

    def hide(self):
        self._canvas_landscape.itemconfig(self.item_tag, state='hidden')
        self._canvas_landscape.itemconfig(self._item_label, state='hidden')

    def setup_ok(self):
        self._connection_effect.setup_ok()

    def setup_nok(self):
        self._connection_effect.setup_nok()

    def dim(self):
        pass


class StatusBannerCanvasItem(CanvasItem):
    """Creates a status banner."""

    def __init__(self, main_window):
        """Instantiates the StatusBannerCanvasItem object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.

        """
        super().__init__(main_window)
        self._tag_text_item = f"{self.item_tag}-1"
        self._tag_box_item = f"{self.item_tag}-2"
        self._text_effect = Effect(main_window, self._tag_text_item, 'fill')
        self._box_effect = Effect(main_window, self._tag_box_item, 'outline')

    def create(self):
        self._create_text()
        self._create_box()

    def _create_text(self):
        width = self._canvas_status.winfo_width()
        pos_x = width / 2
        status_height = self._canvas_status.winfo_height()
        pos_y = status_height / 2
        text = 'NO TUNNEL'.center(30)
        self._canvas_status.create_text(pos_x,
                                        pos_y,
                                        text=text,
                                        font=('', 20, 'normal'),
                                        # anchor=N,
                                        fill=NON_OPERATION,
                                        tags=self._tag_text_item)

    def _create_box(self):
        ax1, ay1, ax2, ay2 = self._canvas_status.bbox(self._tag_text_item)
        ax1 -= 20  # horizontal padding
        ax2 += 20  # horizontal padding
        ay1 -= 10  # horizontal padding
        ay2 += 10  # horizontal padding
        self._canvas_status.create_rectangle(ax1,
                                             ay1,
                                             ax2,
                                             ay2,
                                             outline=NON_OPERATION,
                                             width=THICKNESS_PACKET,
                                             tags=self._tag_box_item)

    def _show_text(self, state=None):
        if state is None:
            self._canvas_status.itemconfig(self._tag_text_item, state='normal')
            self._text_effect.flicker()
        elif state == 'opened':
            text = 'TUNNEL OPENED'.center(30)
            self._canvas_status.itemconfig(self._tag_text_item, text=text, fill=OK_COLOUR)
        elif state == 'broken':
            text = 'TUNNEL BROKEN'.center(30)
            self._canvas_status.itemconfig(self._tag_text_item, text=text, fill=NOK_COLOUR)
        elif state == 'restored':
            text = 'TUNNEL RESTORED'.center(30)
            self._canvas_status.itemconfig(self._tag_text_item, text=text, fill=OK_COLOUR)

    def _show_box(self, state=None):
        if state is None:
            self._canvas_status.itemconfig(self._tag_box_item, state='normal')
            self._box_effect.flicker()
        if state == 'opened':
            self._canvas_status.itemconfig(self._tag_box_item, outline=OK_COLOUR)
        elif state == 'broken':
            self._canvas_status.itemconfig(self._tag_box_item, outline=NOK_COLOUR)
        elif state == 'restored':
            self._canvas_status.itemconfig(self._tag_box_item, outline=OK_COLOUR)

    def show(self, state=None):  # fix pylint eror: arguments-differ / Parameters differ from overridden 'show' method
        self._show_text(state)
        self._show_box(state)

    def hide(self):
        self._canvas_status.itemconfig(self._tag_text_item, state='hidden')
        self._canvas_status.itemconfig(self._tag_box_item, state='hidden')

    def setup_ok(self):
        self._text_effect.setup_ok()
        self._box_effect.setup_ok()

    def setup_nok(self):
        pass

    def dim(self):
        text = 'TUNNEL CLOSED'.center(26)
        self._canvas_status.itemconfig(self._tag_text_item, text=text)
        self._text_effect.dim()
        self._box_effect.dim()


class PacketCanvasItem(CanvasItem):
    """Creates a visualised TCP packet."""

    def __init__(self, main_window, connection_canvas_items):
        """Instantiates the PacketCanvasItem object.

        Args:
            main_window (MainWindow): An instantiated MainWindow object.
            connection_canvas_items (list): Two instantiated CanvasItems, each representing a Connection.

        """
        super().__init__(main_window)
        self.connection_calculator = ConnectionCalculator(main_window,
                                                          connection_canvas_items[0],
                                                          connection_canvas_items[-1])
        self.total_width = WIDTH_PACKET + (2 * THICKNESS_LINE_PACKET)
        self.start_pos_x = self.connection_calculator.get_x_pos_connection_left_side()
        self.pos_y = self.connection_calculator.get_y_pos_connection()
        self.distance = self.connection_calculator.get_connection_length_outer()

    def create(self):
        self._canvas_landscape.create_rectangle(self.start_pos_x,
                                                self.pos_y - 4,
                                                self.start_pos_x + WIDTH_PACKET,
                                                self.pos_y + 4,
                                                fill=PACKET_FILL_COLOUR,
                                                outline=PACKET_OUTLINE_COLOUR,
                                                width=THICKNESS_LINE_PACKET,
                                                # state='hidden',
                                                tags=self.item_tag)

    def show(self):
        self._canvas_landscape.itemconfig(self.item_tag, state='normal')

    def hide(self):
        self._canvas_landscape.itemconfig(self.item_tag, state='hidden')

    def setup_ok(self):
        pass

    def setup_nok(self):
        pass

    def dim(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: powermolegui.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for powermolegui.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import logging.config
from powermolegui.lib.logging import LoggingHandler
from powermolegui.lib.windows import MainWindow

__author__ = '''Vincent Schouten <powermole@protonmail.com>'''
__docformat__ = '''google'''
__date__ = '''08-10-2020'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<powermole@protonmail.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".


def main():
    """Main method.

    This method holds what you want to execute when
    the script is run on command line.
    """
    logger = logging.getLogger()  # Returns a logger (which enables log messages to be printed to the terminal)
    logger.setLevel(logging.DEBUG)
    main_window = MainWindow()
    logging_win_handler = LoggingHandler(main_window)  # Handlers send the log records (created by loggers) to the GUI
    logger.addHandler(logging_win_handler)
    main_window.logging_win_handler = logging_win_handler
    main_window.mainloop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: windows.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for tkinter windows.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import os
import inspect
import threading
import signal
from dataclasses import dataclass
import tkinter as tk
from tkinter import DISABLED, NORMAL
from tkinter.filedialog import askopenfilename
from powermolegui.lib.logging import LoggerMixin
from powermolegui.lib.application import application
from powermolegui.lib.frames import MainFrame, CommandFrame
from powermolegui.lib.helpers import ItemsGenerator, parse_configuration_file

__author__ = '''Vincent Schouten <powermole@protonmail.com>'''
__docformat__ = '''google'''
__date__ = '''08-10-2020'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<powermole@protonmail.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# Constants regarding powermole window for non-retina (generic) screens
WINDOW_WIDTH = 700
WINDOW_HEIGHT = 500

# Constants regarding dimensions of non-retina (generic) screen
GENERIC_SCREEN_WIDTH = 1920


def determine_scale(screen_width):
    """Sets the width of the application screen depending on type of screen."""
    if screen_width <= GENERIC_SCREEN_WIDTH:
        scale = 1
    else:  # retina screen
        scale = 2
    return scale


@dataclass
class CollectionMenuBars:
    """Holds the menu bars."""

    file_menu: tk.Menu
    execution_menu: tk.Menu
    send_menu: tk.Menu
    logging_menu: tk.Menu
    quit_menu: tk.Menu


class MainWindow(tk.Tk, LoggerMixin):  # pylint: disable=too-many-instance-attributes
    """Represents the main window of an application.

    In an Tkinter application, the instance of the Tk class represents the main window.
    """

    def __init__(self, *args, **kwargs):
        """Instantiates the main window object."""
        tk.Tk.__init__(self, *args, **kwargs)
        LoggerMixin.__init__(self)
        signal.signal(signal.SIGINT, self._signal_handler)
        self.scale = 0
        self._script_path = self._determine_script_path()
        self._set_title_icon()
        self._set_size_window()
        self._query_windowingsystem()
        self._menu_bars = None  # holds the Tk menu bars
        self._create_menu()
        self._bind_to_event()
        self._path_config_file = None  # holds the path to the configuration file; set by _config_file_dialog()
        self.protocol("WM_DELETE_WINDOW", self.close_window)
        self.main_frame = MainFrame(self, self.scale)
        self.main_frame.pack(side="top", fill="both", expand=True)
        self.main_frame.config(highlightthickness=2)
        self._set_scrollregion(init=True)
        self.logging_win_handler = None  # holds the (instantiated) logger handler
        self.command_window = None  # holds a TopLevel widget object proving interface for user for sending commands
        self.transfer_window = None  # IN DEVELOPMENT
        self.should_terminate_application = False  # if True, the application will stop running and clean up
        self.configuration = None  # holds configuration parameters; set by show_config_graphics()
        self.canvas_items = None  # holds canvas items; generated by (Canvas)ItemGenerator
        self.instructor = None  # holds *Instructor object which provides 'send command' and 'send file' functions

    def _signal_handler(self, signum, frame):  # pylint: disable=unused-argument
        """Handles the SIGINT generated by an user when hitting Ctrl+C in the shell."""
        self.stop_application()

    def _determine_script_path(self):  # pylint: disable=no-self-use
        running_script = inspect.getframeinfo(inspect.currentframe()).filename  # /powermolegui/lib/windows.py
        running_script_dir = os.path.dirname(os.path.abspath(running_script))
        return os.path.dirname(running_script_dir)  # /powermolegui

    def _set_title_icon(self):
        self.title("powermole")
        path_file = os.path.join(self._script_path, 'icon', 'application_icon_tunnel.png')
        # https://stackoverflow.com/questions/11176638/tkinter-tclerror-error-reading-bitmap-file
        img = tk.PhotoImage(file=path_file)
        self.iconphoto(True, img)

    def _set_size_window(self):
        screen_width = self.winfo_screenwidth()  # width of the computer screen
        screen_height = self.winfo_screenheight()  # height of the computer screen
        self.scale = determine_scale(screen_width)  # IN DEVELOPMENT
        win_width = WINDOW_WIDTH * self.scale  # width of the main window
        win_height = WINDOW_HEIGHT * self.scale  # height of the main window
        start_x = (screen_width / 2) - (win_width / 2)
        start_y = (screen_height / 2) - (win_height / 2)
        self.geometry(f'{win_width}x{win_height}+{int(start_x)}+{int(start_y)}')
        self.resizable(True, True)
        # self._logger.info("screen size is: %s x %s", (ws, hs))  # can't work as logger is instantiated later
        print(f"screen size is: {screen_width} x {screen_height}")
        print(f"window size: {win_width} x {win_height}")

    def _query_windowingsystem(self):
        print(f"windowing system: {self.tk.call('tk', 'windowingsystem')}")

    def _create_menu(self):
        self.option_add('*tearOff', False)
        menubar = tk.Menu(self)
        file_menu = tk.Menu(menubar)
        execution_menu = tk.Menu(menubar)
        send_menu = tk.Menu(menubar)
        logging_menu = tk.Menu(menubar)
        quit_menu = tk.Menu(menubar)
        self._menu_bars = CollectionMenuBars(file_menu, execution_menu, send_menu, logging_menu, quit_menu)

        file_menu.add_command(label='Open', command=self.config_file_dialog)
        file_menu.entryconfig('Open', accelerator='Ctrl+O', state=NORMAL)
        file_menu.add_command(label='Open Recent', command=self.retrieve_recently_opened)
        file_menu.entryconfig('Open Recent', accelerator='Ctrl+T', state=NORMAL)
        menubar.add_cascade(label='File', menu=file_menu)

        execution_menu.add_command(label='Run Application', command=self.run_application)
        execution_menu.entryconfig('Run Application', accelerator='Ctrl+R', state=DISABLED)
        execution_menu.add_command(label='Stop Application', command=self.stop_application)
        execution_menu.entryconfig('Stop Application', accelerator='Ctrl+C', state=DISABLED)
        menubar.add_cascade(label='Execution', menu=execution_menu)

        send_menu.add_command(label='Send File', command=None)
        send_menu.entryconfig('Send File', accelerator='Ctrl+F', state=DISABLED)
        send_menu.add_command(label='Send Command', command=self.open_command_window)
        send_menu.entryconfig('Send Command', accelerator='Ctrl+M', state=DISABLED)
        menubar.add_cascade(label='Send', menu=send_menu)

        var_info = tk.BooleanVar()
        var_debug = tk.BooleanVar()
        logging_menu.add_checkbutton(label='Info', onvalue=1, offvalue=0, variable=var_info)
        logging_menu.add_checkbutton(label='Debug', onvalue=1, offvalue=0, variable=var_debug)
        menubar.add_cascade(label='Logging', menu=logging_menu)
        var_info.set(value=1)
        logging_menu.entryconfig('Info', state=DISABLED)
        logging_menu.entryconfig('Debug', state=DISABLED)
        # logging_win_handler.setLevel(......)  # IN DEVELOPMENT (including the code block above)

        quit_menu.add_command(label='Quit', command=self.close_window)
        quit_menu.entryconfig('Quit', accelerator='Ctrl+Q')
        menubar.add_cascade(label='Quit', menu=quit_menu)

        self.config(menu=menubar)  # self == parent = tk.Tk()

    def change_state_menu_bar_entry(self, entry, label, state):
        """Changes the state of the menu bar."""
        if entry == 'send':
            self._menu_bars.send_menu.entryconfig(label, state=state)  # the 'Send File' menu entry is in development
        elif entry == 'file':
            self._menu_bars.file_menu.entryconfig(label, state=state)
        elif entry == 'execution':
            self._menu_bars.execution_menu.entryconfig(label, state=state)

    def _bind_to_event(self):
        self.bind('<Control-o>', lambda e: self.config_file_dialog())
        self.bind('<Control-t>', lambda e: self.retrieve_recently_opened())
        self.bind('<Control-r>', lambda e: self.run_application())
        self.bind('<Control-c>', lambda e: self.stop_application())
        self.bind('<Control-f>', lambda e: None)  # IN DEVELOPMENT
        self.bind('<Control-m>', lambda e: self.open_command_window())
        self.bind('<Control-q>', lambda e: self.close_window())

    def open_command_window(self):
        """Opens interface for the user to send commands to last host and show output."""
        self.command_window = CommandWindow(self)

    def _set_scrollregion(self, init=False):
        """Sets a scroll region that encompasses all the canvas items."""
        self.main_frame.canvas_frame.canvas_landscape.update_idletasks()
        w_height = self.main_frame.canvas_frame.canvas_landscape.winfo_height()
        w_width = self.main_frame.canvas_frame.canvas_landscape.winfo_width()
        if init:
            self.main_frame.canvas_frame.canvas_landscape.config(scrollregion=(0, 0, w_width, w_height))
        else:
            # retrieve the x-axis at the far right of the last drawn item:
            _, _, x_axis_2, _ = self.main_frame.canvas_frame.canvas_landscape.bbox('all')
            if x_axis_2 <= w_width:  # if the bounding box of all items is smaller than the canvas width,
                # dismiss bounding box size
                self.main_frame.canvas_frame.canvas_landscape.config(scrollregion=(0, 0, w_width, w_height))
            else:
                self.main_frame.canvas_frame.canvas_landscape.config(scrollregion=(0, 0, x_axis_2 + 100, w_height))

    def retrieve_recently_opened(self):  # fix this py_lint error --> inconsistent-return-statements
        """Retrieves the recently opened configuration file stored in /settings."""
        path_file_recent = os.path.join(self._script_path, 'settings', 'recently_opened_config_file')
        try:
            with open(path_file_recent) as file:
                self._path_config_file = file.read().rstrip()
        except FileNotFoundError:
            pass
        config_thread = threading.Thread(target=self._show_config_graphics)
        config_thread.start()

    def _write_to_recently_opened(self, path_config_file):
        """Stores the location to the recently opened configuration file."""
        path_file_recent = os.path.join(self._script_path, 'settings', 'recently_opened_config_file')
        with open(path_file_recent, 'w') as file:
            file.write(path_config_file)

    def config_file_dialog(self):
        """Shows the file dialog."""
        self.main_frame.canvas_frame.canvas_landscape.delete("all")
        file_types = [('powermole config file', '*.json')]
        self._path_config_file = askopenfilename(filetypes=file_types)
        if self._path_config_file:
            config_thread = threading.Thread(target=self._show_config_graphics)
            config_thread.start()
            self._write_to_recently_opened(self._path_config_file)

    def _show_config_graphics(self):
        """Creates canvas items and shows the landscape based on the config file.

        This method is called by _config_file_dialog when
        the user opens ("Open") a powermole configuration file.
        """
        if self._path_config_file:  # return True if the variable is set with a path
            self.configuration = parse_configuration_file(self._path_config_file)  # return configuration object
            if self.configuration:
                items_generator = ItemsGenerator(self, self.configuration)
                self.canvas_items = items_generator.create_canvas_items()  # creates all canvas items
                items_generator.show_landscape(self.canvas_items)
                self._set_scrollregion()
                self.change_state_menu_bar_entry('execution', 'Run Application',
                                                 NORMAL)  # enable menu bar to start/stop application

    def run_application(self):
        """Starts the application."""
        run_thread = threading.Thread(target=application, args=(self,), name='running_application')
        run_thread.start()
        self.change_state_menu_bar_entry('file', 'Open', DISABLED)
        self.change_state_menu_bar_entry('file', 'Open Recent', DISABLED)
        self.change_state_menu_bar_entry('execution', 'Run Application', DISABLED)
        self.change_state_menu_bar_entry('execution', 'Stop Application', NORMAL)

    def stop_application(self):
        """Sets the should_terminate var to True.

        The application is running indefinitely until the user
        hits ctrl + c in the window. The window widget will
        capture the event, and sets the var should_terminate to True.
        The application, polling this var, will break the loop,
        and dismantles the tunnel.
        """
        self.should_terminate_application = True

    def close_window(self):
        """Closes the window."""

        def _is_thread_application_running():
            result = False
            for thread in threading.enumerate():
                if thread.name == 'running_application':
                    result = True
            return result

        if _is_thread_application_running():
            self._logger.info('*** window _cannot_ be closed as Tunnel is operational (press Ctrl+c) ***')
        else:
            self.destroy()


class CommandWindow(tk.Toplevel):
    """Represents an interface for the user to send commands to destination host and show output."""

    def __init__(self, parent, *args, **kwargs):
        """Instantiates the TopLevel object."""
        super().__init__(*args, **kwargs)
        self._bind_to_event()
        self._is_return_pressed = False
        self._set_size()
        self.instructor = parent.instructor
        self.scale = 0
        self.title("Interface")
        self.protocol("WM_DELETE_WINDOW", self.close_window)
        self.sub_command_window = CommandFrame(self)

    def _bind_to_event(self):
        self.bind("<Return>", lambda e: self.send_command())
        self.bind('<Control-c>', lambda e: self.close_window())

    def _set_size(self):
        screen_width = self.winfo_screenwidth()  # width of the computer screen
        screen_height = self.winfo_screenheight()  # height of the computer screen
        self.scale = determine_scale(screen_width)
        win_width = WINDOW_WIDTH * 0.7  # width of the main window
        win_height = WINDOW_HEIGHT * 0.7  # height of the main window
        start_x = (screen_width / 2) - (win_width / 2)
        start_y = (screen_height / 2) - (win_height / 2)
        self.geometry('%dx%d+%d+%d' % (win_width, win_height, start_x, start_y))  # geometry behaves erratic!
        self.resizable(True, True)

    def close_window(self):
        """Closes this top level widget."""
        self.destroy()

    def send_command(self):
        """Captures the input of the user and shows output returned from *Instructor."""
        input_ = self.sub_command_window.command_entry.entry.get()
        self.sub_command_window.command_entry.entry.delete(0, 'end')
        self._is_return_pressed = False
        output = self.instructor.send_command(input_)
        self._parse_output(output)

    def _parse_output(self, output):
        output_str = output.decode("utf-8")
        output_line = output_str.split('\n')
        for line in output_line:
            self.sub_command_window.command_response.text.insert('end', line + '\n')
            self.sub_command_window.command_response.text.see("end")
        self.sub_command_window.command_response.text.insert('end', '\n')


class TransferWindow(tk.Toplevel):
    """Represents an interface for the user to select files locally to be copied to destination host.

    IN DEVELOPMENT!
    """

    def __init__(self, *args, **kwargs):
        """Instantiates the TopLevel object."""
        super().__init__(*args, **kwargs)  # with super(), no self as argument is needed
        self.scale = 0
        self.title("Interface")
        self._set_size()

    def _set_size(self):
        screen_width = self.winfo_screenwidth()  # width of the computer screen
        screen_height = self.winfo_screenheight()  # height of the computer screen
        self.scale = determine_scale(screen_width)
        win_width = WINDOW_WIDTH * 0.7  # width of the main window
        win_height = WINDOW_HEIGHT * 0.7  # height of the main window
        start_x = (screen_width / 2) - (win_width / 2)
        start_y = (screen_height / 2) - (win_height / 2)
        self.geometry('%dx%d+%d+%d' % (win_width, win_height, start_x, start_y))  # geometry behaves erratic!
        self.resizable(True, True)
